### CONFIGURAÇÕES BÁSICAS PARA RODAR TODAS AS ROTINAS ###

import yaml
from getpass import getuser
from stpstone.cals.handling_dates import DatesBR
from stpstone.loggs.create_logs import CreateLog


def reading_yaml(yaml_path):
    '''
    DOCSTRING: ABRIR ARQUIVOS .YAML
    INPUTS: CAMINHO
    OUTPUTS: ATRIBUIÇÃO DE INPUTS 
    '''
    return yaml.load(open(yaml_path), Loader=yaml.FullLoader)


def iniciating_logging(logger_name, parent_destination_log=None):
    '''
    DOCSTRING: INICIALIZING LOGGING LIB
    INPUTS: COMPLETE PATH TO LOGGING DIRECTORY
    OUTPUTS: -
    '''
    # creating parent directory
    if parent_destination_log != None:
        dispatch = CreateLog().creating_parent_folder(parent_destination_log)
        CreateLog().infos(logger_name, 'Diretório pai de logs: {}'.format(parent_destination_log))
        if dispatch == 'OK':
            CreateLog().infos(logger_name, 'Diretório pai criado com sucesso.')
        elif dispatch == 'NOK':
            CreateLog().infos(
                logger_name, 'Diretório pai já havia sido criado, nenhuma medida foi tomada.')
        else:
            raise Exception(
                'O valor retornado na função deve ser OK ou NOK para criação do diretório')
    # iniciating routine
    CreateLog().infos(logger_name, 'Rotina iniciada em {}'.format(
        str(DatesBR().curr_date_time())))
    CreateLog().infos(logger_name, 'Operador da rotina {}'.format(str(getuser())))



### INTEGRATION WITH GPT LLM ###

from typing import List, Optional
from openai import OpenAI
from openai.api_resources.chat_completion import ChatCompletion
from stpstone.handling_data.dicts import HandlingDicts
from stpstone.settings._global_slots import YAML_LLMS


class GPT():

    def __init__(self, api_key:str, str_model:str, int_max_tokens:int=100, 
                 str_context:Optional[str]=None, bl_stream:bool=False) -> None:
        '''
        REFERENCES:
            - DOCUMENTATION: https://platform.openai.com/docs/guides/gpt
            - MODELS AVAILABLE: https://platform.openai.com/docs/models/gp
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        self.api_key = api_key
        self.str_model = str_model
        self.int_max_tokens = int_max_tokens
        self.str_context = str_context
        self.bl_strem=bl_stream,
        self.client = OpenAI(api_key=self.api_key)

    def run_prompt(self, list_tuple:List[tuple]) -> ChatCompletion:
        '''
        DOCSTRING: RUN LLM WITH PROMPT
        INPUTS: 
            - LIST_TUPLE:LIST[TUPLE] IN ORDER TO CREATE A PROMPT TO REQUEST THE LLM
        OUTPUTS: STR
        '''
        # setting variables
        list_ = list()
        # user's information in order to build the prompt
        dict_content = {
            YAML_LLMS['openai']['key_role']: YAML_LLMS['openai']['value_user']
        }
        # looping within types and messages, in order to create the content of the prompt
        for tup_ in list_tuple:
            if tup_[0] == 'text':
                list_.append({
                    'type': str(tup_[0]).lower(),
                    str(tup_[0]).lower(): str(tup_[1])
                })
            elif tup_[0] == 'image_url':
                list_.append({
                    'type': str(tup_[0]).lower(),
                    'image_url': {
                        'url': str(tup_[1]).lower()
                    }
                })
        # creating the message content
        dict_content = HandlingDicts().merge_n_dicts(dict_content, {'content': list_})
        # add context info, regarding the prompt
        if self.str_context is not None:
            list_prompt = [
                {
                    'role': 'system', 
                    'content': self.str_context
                }, 
                dict_content
            ]
        else:
            list_prompt = [dict_content]
        # request llm info
        return self.client.chat.completions.create(
            model=self.str_model,
            messages=list_prompt,
            max_tokens=self.int_max_tokens,
            strem=self.bl_strem
        )



import pandas as pd
import typeform as tf
from stpstone.handling_data.json_format import JsonFiles


class DealingTypeForm:

    def downloading_response(self, json_file, api_token, form_id, since_date=None, until_date=None):
        '''
        DOCSTRING: DOWNLOAD RESPONSES OF A TYPEFORM TO A NETWORK PATH IN JSON EXTENSION
        INPUTS: JSON FILE (STR), API TOKEN (STR), FORM_ID (STR), SINCE DATE AND UNTIL DATE FORMAT:
        'YYYY-MM-DDT00:00:00.00Z', AS WELL A STR TYPE
        OUTPUTS: 'OK' OR 'NOK' FOR SUCCESSFUL SAVING
        '''

        # fetching data
        form_mem = tf.Typeform(token=api_token).responses.list(uid=form_id,
                                                               since=since_date,
                                                               until=until_date)

        # returning status of saving
        return JsonFiles().dump_message(form_mem, json_file)

    def dictionary_ref_questions(self, api_token, form_id):

        # placing an empty dict in memory
        dict_ref_questions = dict()

        # fetching data of form of interest
        tf_infos = tf.Typeform(token=api_token).forms.get(form_id)

        # building a dict: key - ref / value - question
        dict_ref_questions['response_id'] = '#'

        for field in tf_infos['fields']:
            dict_ref_questions[field['ref']] = field['title']

        dict_ref_questions['landed'] = 'Start Date (UTC)'
        dict_ref_questions['submitted_at'] = 'Submit Date (UTC)'
        dict_ref_questions['network_id'] = 'Network ID'

        return dict_ref_questions

    def response_to_dataframe(self, ref_questions_translator, responses_json_path,
                              api_token, form_id):
        '''
        DOCSTRING: RETURN PANDAS DATAFRAME OF RESPONSE TYPEFORM 
        INPUTS: DICT OF REFS KEYS X QUESTIONS VALUES, 
        RESPONSE PATH (OUGHT BE DOWNLOADED TO NETWORK PATH), API TOKEN (STR) AND FORM ID (STR)
        OUTPUTS: LIST OF PILED RESPONSES
        '''

        # import json to memory
        responses_json_mem = JsonFiles().load_message(responses_json_path)

        # creating memory slots
        list_piled_responses = list()
        dict_current_responses = dict()

        # aggregate dicts in a list
        for item in responses_json_mem['items']:
            for _, value in ref_questions_translator.items():
                dict_current_responses[value] = None
            if item['response_id']:
                dict_current_responses['#'] = item['response_id']
            if item['landed_at']:
                dict_current_responses['Start Date (UTC)'] = item['landed_at']
            if item['submitted_at']:
                dict_current_responses['Submit Date (UTC)'] = item['submitted_at']
            if item['metadata']['network_id']:
                dict_current_responses['Network ID'] = item['metadata']['network_id']
            for answer in item['answers']:
                if answer['type'] == 'choices':
                    dict_current_responses[ref_questions_translator[
                        answer['field']['ref']]] = answer[
                        answer['type']]['labels']
                else:
                    dict_current_responses[ref_questions_translator[
                        answer['field']['ref']]] = answer[answer['type']]
            list_piled_responses.append(dict_current_responses.copy())

        # exporting list to dataframe
        return pd.DataFrame(list_piled_responses)



### INTEGRATION TO BANK APIS ###

import pandas as pd
from pynubank import Nubank, MockHttpClient
from stpstone.handling_data.json import JsonFiles
from stpstone.handling_data.pd import DealingPd
from stpstone.cals.handling_dates import DatesBR


class NubankIntegration:

    def authentification(self, individual_register_number, password, complete_path_cert_p12=None,
                         auth_method='qr_code', mock=True, uuid_mock='123'):
        '''
        REFERENCES: https://github.com/andreroggeri/pynubank/blob/master/examples/login-qrcode.md,
            https://github.com/andreroggeri/pynubank/blob/master/examples/login-certificate.md,
            https://github.com/andreroggeri/pynubank
        DOCSTRING: LOGIN WITH QR CODE
        INPUTS: INDIVIDUAL REGISTER NUMBER (CPF), PASSWORD, COMPLETE PATH CERTIFICATION P12 (
            LOGGING WITH CERTIFICATION TO COMMUNICATE WITH NUBANK SERVER, NONE AS DEFAULT),
            AUTHENTICATION METHOD, MOCK (ENTERING MODE, AS A BOOLEAN), UUID_MOCK (123 AS DEFAULT)
        OUTPUTS: NUBANK OBJECT
        '''
        # wheter or not log through mock authentication
        if mock == True:
            # nubank object
            nu = Nubank(MockHttpClient())
            # authenticate credentials
            nu.authenticate_with_qr_code(
                individual_register_number, password, uuid_mock)
            # return access to nubank objects, granted through authentification
            return nu
        elif mock == False:
            # nubank object
            nu = Nubank()
            # validating method of logging
            if auth_method == 'qr_code':
                # after this command will be displayed a QRCode in a console, please scan through
                #   the app to grante entrance; this is available in the menu: NU > Perfil > Acesso
                #   pelo site, in the brazillian portugues translation
                uuid, qr_code = nu.get_qr_code()
                qr_code.print_ascii(invert=True)
                input('After scanning, please hit any key to continue...')
                # authenticate credentials
                nu.authenticate_with_qr_code(
                    individual_register_number, password, uuid)
                # return access to nubank objects, granted through authentification
                return nu
            elif auth_method == 'certification':
                nu.authenticate_with_cert(
                    individual_register_number, password, complete_path_cert_p12)
                # return access to nubank objects, granted through authentification
                return nu
            else:
                raise Exception(
                    'Please consider the qr code or certification method to login')
        else:
            raise Exception('Authetication mock variable ouught be a boolean')

    def credit_card_bill(self, nu, bl_return_pd=True, float_den_amount=100.0,
                         col_id='ID', col_description='DESCRIPTION', col_category='CATEGORY',
                         col_subcategory='SUBCATEGORY', col_amount='AMOUNT',
                         col_amount_wo_iof='AMOUNT_WITHOUT_IOF', col_time='TIME',
                         col_source='SOURCE', col_title='TITLE', col_account='ACCOUNT',
                         col_status='STATUS', col_href_prod_global='HREF_PROD_GLOBAL',
                         bl_tokenized='BL_TOKENIZED', col_href_transaction='HREF_TRANSACTION',
                         col_lat='LAT', col_lon='LON'):
        '''
        REFERENCES: https://github.com/andreroggeri/pynubank
        DOCSTRING: NUBANK CREDIT CARD BILL
        INPUTS: NU OBJECT, RETURNED THROUGH AUTHENTIFICATION
        OUTPUTS: JSON (DICT WITH CARD STATEMENTS, BILLS AND BILLS DETAILS AS KEYS)
        '''
        # setting variables
        list_ser = list()
        # import dictionary with card statements
        list_ser_card_statements = nu.get_card_statements()
        # return pandas dataframe, if it is user's will
        if bl_return_pd == True:
            #   looping through card statements aiming to build a pandas dataframe
            for dict_ in list_ser_card_statements:
                #   dealing with eventually missing data
                if 'subcategory' in dict_['details']:
                    subcategory = dict_['details']['subcategory'].upper()
                else:
                    subcategory = 'N/A'
                if 'source' in dict_:
                    source = dict_['source'].upper()
                else:
                    source = 'N/A'
                if 'account' in dict_:
                    account = dict_['account'].upper()
                else:
                    account = 'N/A'
                if 'status' in dict_:
                    status = dict_['details']['status'].upper()
                else:
                    status = 'N/A'
                if 'tokenized' in dict_:
                    tokenized = str(dict_['tokenized']).upper()
                else:
                    tokenized = 'N/A'
                if 'href' in dict_:
                    href_transaction = dict_['href']
                else:
                    href_transaction = 'N/A'
                if 'amount_without_iof' in dict_:
                    amount_wo_iof = dict_['amount_without_iof']
                else:
                    amount_wo_iof = dict_['amount']
                if 'lat' in dict_['details']:
                    lat = dict_['details']['lat']
                else:
                    lat = 99999999
                if 'lon' in dict_['details']:
                    lon = dict_['details']['lon']
                else:
                    lon = 99999999
                #   serialized data to pandas dataframe
                list_ser.append({
                    col_id: dict_['id'].upper(),
                    col_description: dict_['description'].upper(),
                    col_category: dict_['category'].upper(),
                    col_subcategory: subcategory,
                    col_amount: dict_['amount'] / float_den_amount,
                    col_amount_wo_iof: amount_wo_iof / float_den_amount,
                    col_time: dict_['time'],
                    col_source: source,
                    col_title: dict_['title'].upper(),
                    col_account: account,
                    col_status: status,
                    col_href_prod_global: dict_['_links']['self']['href'],
                    bl_tokenized: tokenized,
                    col_href_transaction: href_transaction,
                    col_lat: lat,
                    col_lon: lon
                })
            #   pandas dataframe
            df_card_statements = pd.DataFrame(list_ser)
            #   changing column types
            df_card_statements = df_card_statements.astype({
                col_id: str,
                col_description: str,
                col_category: str,
                col_subcategory: str,
                col_amount: float,
                col_amount_wo_iof: float,
                col_time: str,
                col_source: str,
                col_title: str,
                col_account: str,
                col_status: str,
                col_href_prod_global: str,
                bl_tokenized: str,
                col_href_transaction: str,
                col_lat: float,
                col_lon: float
            })
            #   datetime colum types
            df_card_statements[col_time] = pd.to_datetime(
                df_card_statements[col_time], errors='ignore')
            #   sort values according to date and category
            df_card_statements.sort_values([col_time, col_category], ascending=[True, True],
                                           inplace=True)
            #   return pandas dataframe
            return df_card_statements
        # if raise an error in bills, return just card statements
        try:
            # list of dicts with all credit card bills
            bills = nu.get_bills()
            # get credit card bill details
            bills_details = [nu.get_bill_details(
                bills[i]) for i in range(1, len(bills))]
            # send nubank credit card message
            jsonify_message = {
                'card_statements': list_ser_card_statements,
                'bills': bills,
                'bills_details': bills_details
            }
            return JsonFiles().send_json(jsonify_message)
        except:
            return list_ser_card_statements

    def current_account(self, nu):
        '''
        REFERENCES: https://github.com/andreroggeri/pynubank
        DOCSTRING: NUBANK CURRENT ACCOUNT DETAILS
        INPUTS: NU OBJECT, RETURNED THROUGH AUTHENTIFICATION
        OUTPUTS: JSON (DICT WITH CARD STATEMENTS, BILLS AND BILLS DETAILS AS KEYS)
        '''
        jsonify_message = {
            'account_statements': nu.get_account_statements(),
            'current_account': nu.get_account_balance()
        }
        return JsonFiles().send_json(jsonify_message)



### HANDLING XML FILES ###

import xml.etree.ElementTree as et
from bs4 import BeautifulSoup


class XMLFiles:

    def fetch_et(self, path_xml):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # import xtree to memory
        xtree = et.parse(path_xml)
        # return root
        return xtree.getroot()

    def get_attrib_node_et(self, node, attrib_name):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        return node.attrib.get(attrib_name)

    def xml_parser(self, xml_file):
        '''
        DOCSTRING: XML PARSER THROUGH BEAUTIFULSOUP
        INPUTS: XML FILE COMPLETE PATH
        OUTPUTS: SOUP
        '''
        infile = open(xml_file, 'r', encoding='UTF-8')
        contents = infile.read()
        return BeautifulSoup(contents, 'xml')

    def xml_memory_parser(self, xml_cache):
        '''
        DOCSTRING: FETCH XML
        INPUTS: XML_CACHE
        OUTPUTS: XML IN MEMORY - SOUP
        '''
        return BeautifulSoup(xml_cache, 'xml')

    def xml_find(self, soup_xml, tag):
        '''
        DOCSTRING: GET SOUP ELEMENT WITHIN A TAG
        INPUTS: SOUP_XML, TAG
        OUTPUTS: XML ELEMENT
        '''
        return soup_xml.find(tag)

    def xml_find_all(self, soup_xml, tag):
        '''
        DOCSTRING: GET A LIST OF ELEMENTS IN SOUP XML THAT RESPECT A GIVEN TAG
        INPUTS: SOUP_XML, TAG
        OUTPUTS: XML ELEMENT        
        '''
        return soup_xml.find_all(tag)

    def xml_get_text(self, soup_xml):
        '''
        DOCSTRING: GET TEXT WITHIN AN XML ELEMENT
        INPUTS: SOUP_XML
        OUTPUTS: STRING
        '''
        return soup_xml.get_text()



### CRIACAO DE LOGS ###

import logging
import os
import time


class CreateLog:
    '''
    DOCSTRING: PREENCHIMENTO DE LOGS AO LONGO DA ROTINA
    INPUTS: -
    OUTPUTS: INFOS
    '''

    def creating_parent_folder(self, new_path):
        '''
        DOCSTRING: CRIAÇÃO DE PASTA NA REDE PARA EMBARCAR LOGS
        INPUTS: NEW PATH
        OUTPUTS: INFOS
        '''
        if not os.path.exists(new_path):
            os.makedirs(new_path)
            return 'OK'
        else:
            return 'NOK'

    def basic_conf(self, complete_path=None, basic_level='info'):
        '''
        DOCSTRING: INICIO ROTINA DE FORNECIMENTO DE LOGS
        INPUTS: -
        OUTPUTS: LOGGER
        '''
        # defining basic level of logging
        if basic_level == 'info':
            level = logging.INFO
        elif basic_level == 'debug':
            level = logging.DEBUG
        else:
            raise Exception(
                'Level was not properly defined in basic config of logging, please check')
        # logging message format
        logging.basicConfig(
            level=level,
            filename=complete_path,
            format=(
                '%(asctime)s.%(msecs)03d %(levelname)s {%(module)s} [%(funcName)s] '
                '%(message)s'
            ),
            datefmt='%Y-%m-%d,%H:%M:%S',
        )
        # console stream handler
        console = logging.StreamHandler()
        console.setLevel
        # defining logger object
        logger = logging.getLogger(__name__)
        # return logger
        return logger

    def infos(self, logger, msg_str):
        '''
        DOCSTRING: FORNECIMENTO DE INFORMAÇÕES SOBRE O ANDAMENTO DO CÓDIGO
        INPUTS: MENSAGEM EM STRING
        OUTPUTS: INFO
        '''
        return logger.info(msg_str)

    def warnings(self, logger, msg_str):
        '''
        DOCSTRING: ALERTAS PARA PONTOS DE ATENÇÃO NO CÓDIGO
        INPUTS: MENSAGEM EM STRING
        OUTPUTS: WARNING
        '''
        return logger.warning(msg_str)

    def errors(self, logger, msg_str):
        '''
        DOCSTRING: FALHAS POR PARTE DO USUÁRIO OU CONCEPÇÃO DO CÓDIGO
        INPUTS: MENSAGEM EM STRING
        OUTPUTS: ERROR
        '''
        return logger.error(msg_str, exc_info=True)

    def critical(self, logger, msg_str):
        '''
        DOCSTRING: APONTAR ERROS CRÍTICOS NO CÓDIGO
        INPUTS: MENSAGEM EM STRING
        OUTPUTS: ERROR
        '''
        return logger.error(msg_str)


# decorators
def timeit(method):
    '''
    REFERENCES: https://medium.com/pythonhive/python-decorator-to-measure-the-execution-time-of-methods-fa04cb6bb36d
    DOCSTRING: TIMING DECORRATOR TO MEASURE ELAPSED TIME TO EXECUTE A FUNCTION
    INPUTS: -
    OUTPUTS: ELAPSED TIME PRINTED
    '''
    def timed(*args, **kw):
        ts = time.time()
        result = method(*args, **kw)
        te = time.time()
        if 'log_time' in kw:
            name = kw.get('log_name', method.__name__.upper())
            kw['log_time'][name] = int((te - ts) * 1000)
        else:
            print('%r  %2.2f ms' %
                  (method.__name__, (te - ts) * 1000))
        return result
    return timed

def conditional_timeit(bl_use_timer):
    '''
    DOCSTRING: APPLIES THE @TIMEIT DECORATOR CONDITIONALLY BASED ON `USE_TIMER`
    INPUT:
        - USE_TIMER: BOOLEAN INDICATING WHETHER TO APPLY TIMING.
    OUTPUT:
        - A FUNCTION WRAPPED WITH THE @TIMEIT DECORATOR IF `USE_TIMER` IS TRUE.
    '''
    def decorator(method):
        if bl_use_timer:
            return timeit(method)
        return method
    return decorator


### API TO CONNECT IN REUTES DATABASE ###

from requests import request
from urllib.parse import urljoin
import sys
sys.path.append(
    r'C:\Users\Guilherme\OneDrive\Dev\Python\Packages')
from stpstone.handling_data.object import HandlingObjects


class Reuters:

    def fetch_data(self, app, payload={}, method='GET',
                   endpoint='https://apiservice.reuters.com/api/'):
        '''
        REFERENCES: https://www.reuters.com/markets/currencies
        DOCSTRING: FETCHING DATA FROM RETURES CHARTS
        INPUTS: ENDPOINT (DEFAULT) AND METHOD (DEFAULT)
        OUTPUTS: JSON (ACCESS TOKEN, TOKEN TYPE AND EXPIRES IN KEYS)
        '''
        # requesting info from reuters api
        return request(
            method=method, url=urljoin(endpoint, app), params=payload, verify=False).text

    def token(self, api_key, deviceid, app='service/modtoken',
              method='GET'):
        '''
        REFERENCES: https://www.reuters.com/markets/currencies
        DOCSTRING: GET TOKEN FROM REUTERS API
        INPUTS: APP (DEFAULT) AND METHOD (DEFAULT)
        OUTPUTS: JSON (ACCESS TOKEN, TOKEN TYPE AND EXPIRES IN KEYS)
        '''
        # parameters for the endpoint
        payload = {
            'method': 'get',
            'format': 'json',
            'callback': 'getChartData',
            'apikey': api_key,
            'deviceid': deviceid
        }
        # requesting info from reuters api
        json_response = self.fetch_data(app, payload)
        # the api returns a byte (b') type of information, then its transformed to a text
        #   representation and used a literal eval to reach a
        return HandlingObjects().literal_eval_data(json_response, 'getChartData(', ')')

    def quotes(self, currency, app='getFetchQuotes/{}',
               endpoint='https://www.reuters.com/companies/api/'):
        '''
        REFERENCES: https://www.reuters.com/markets/currencies
        DOCSTRING: QUOUTES FROM REUTERS
        INPUTS: CURRENCY, APP (DEFAULT) AND ENDPOINT (DEFAULT)
        OUTPUTS: JSON
        '''
        return HandlingObjects().literal_eval_data(self.fetch_data(urljoin(app, currency),
                                                                   endpoint=endpoint))


# api_key = '72461C50B1CEAD3135BA6BDA53B203D3'
# deviceid = 'E7CDD293-9C3A-5AB9-9181-58E1B572DD44'
# currency = 'USDZAR'

# class_retuers = Reuters(api_key, deviceid)
# json_token = class_retuers.token()
# print(class_retuers.quotes(currency))



### STATISTICAL DESCRIPTION ###

import statistics
import numpy as np
from scipy import stats
from sklearn.metrics import mean_squared_error, mean_absolute_error


class StatiscalDescription:

    def statistical_description(self, array_data):
        '''
        DOCSTRING: STATISTICAL DISCRIPTION (NOBS, MIN, MAX, MEAN, VARIANCE, SKEWNESS, KURTOSIS)
        INPUTS: array OF REAL NUMBERS
        OUTPUTS: DICTIONARY
        '''
        return {
            'nobs': stats.describe(np.array(array_data)).nobs,
            'minmax': stats.describe(np.array(array_data)).minmax,
            'mean': stats.describe(np.array(array_data)).mean,
            'median': statistics.median(np.array(array_data)),
            'mode': statistics.mode(np.array(array_data)),
            'variance_sample': stats.describe(np.array(array_data)).variance,
            'standard_deviation_sample': stats.describe(np.array(
                array_data)).variance ** 0.5,
            'skewness': stats.describe(np.array(array_data)).skewness,
            'kurtosis': stats.describe(np.array(array_data)).kurtosis
        }

    def standard_deviation_sample(self, array_data):
        '''
        DOCSTRING: STATISTICAL DEVIATION OF A SAMPLE
        INPUTS: array OF REAL NUMBERS
        OUTPUTS: STATISTICAL DEVIATION - SAMPLE
        '''
        return stats.tstd(np.array(array_data))

    def harmonic_mean(self, array_data):
        '''
        DOCSTRING: HARMONIC MEAN OF A SAMPLE
        INPUTS: array OF REAL NUMBERS
        OUTPUTS: HARMONIC MEAN
        '''
        return statistics.harmonic_mean(np.array(array_data))

    def median_sample(self, array_data):
        '''
        DOCSTRING: MEDIAN OF A SAMPLE
        INPUTS: array OF REAL NUMBERS
        OUTPUTS: MEDIAN
        '''
        return statistics.median(np.array(array_data))

    def mode_sample(self, array_data):
        '''
        DOCSTRING: MODE OF A SAMPLE
        INPUTS: array OF REAL NUMBERS
        OUTPUTS: MODE
        '''
        return statistics.mode(np.array(array_data))

    def covariance(self, array_data_1, array_data_2):
        '''
        REFERENCES: https://stackoverflow.com/questions/48105922/numpy-covariance-between-each-column-of-a-matrix-and-a-array
        DOCSTRING: COVARIANCE OF ARRAY
        INPUTS: TWO ARRAIES '[[]]', NO NEED TO IMPORT ARRAY FUNCTION TO RECOGNISE ITS BEHAVIOUR
        OUTPUTS: COVARIANCE - FLOAT
        '''
        return np.dot(np.array(array_data_2).T - np.array(
            array_data_2).mean(), np.array(array_data_1) - np.array(
                array_data_1).mean(axis=0)) / (np.array(array_data_2).shape[0] - 1)

    def correlation(self, array_data_1, array_data_2):
        '''
        REFERENCES: https://stackoverflow.com/questions/48105922/numpy-covariance-between-each-column-of-a-matrix-and-a-array
        DOCSTRING: TWO ARRAIES '[[]]', NO NEED TO IMPORT ARRAY FUNCTION TO RECOGNISE ITS BEHAVIOUR
        OUTPUTS:
        '''
        return self.covariance(array_data_1, array_data_2) / \
            np.sqrt(np.var(np.array(array_data_2), ddof=1) * np.var(
                np.array(array_data_1), axis=0, ddof=1))

    def mean_squared_error_fitting_precision(self, array_observations, array_predictions):
        '''
        DOCSTRING: MEAN SQUARED ERROR TO TEST FITTING PRECISION
        INPUTS: LIST OF OBSERVATIONS AND LIST OF PREDICTIONS
        OUTPUTS: DICTIONAY WITH MEAN SQUARED ERROR AND ITS SQUARED ROOT
        '''
        float_mean_squared_error = mean_squared_error(array_observations,
                                                      array_predictions)
        return {
            'mean_squared_error': float_mean_squared_error,
            'sqrt_mean_squared_error': np.sqrt(float_mean_squared_error)
        }

    def mean_absolute_error_fitting_precision(self, array_observations, array_predictions):
        '''
        DOCSTRING: MEAN ABSOLUTE ERROR FITTING PRECISION, USED FOR MODELS WITH A PLENTY
            OF OUTLIERS
        INPUTS: LIST OF OBSERVATIONS AND LIST OF PREDICTIONS
        OUTPUTS: DICTIONAY WITH MEAN ABSOLUTE SQUARED ERROR AND ITS SQUARED ROOT
        '''
        float_mean_absolute_squared_error = mean_absolute_error(array_observations,
                                                                array_predictions)
        return {
            'mean_absolute_squared_error': float_mean_absolute_squared_error,
            'sqrt_mean_absolute_squared_error': np.sqrt(float_mean_absolute_squared_error)
        }

    def quantile(self, array_data, q, axis=0):
        '''
        DOCSTRING: QUANTILE
        INPUTS: ARRAY, QUANTILE BETWEEN 0 AND 1, AXIS (0 FOR FLATENNED OR 1 FOR MATRIC)
        OUTPUTS: ARRAY
        '''
        return np.quantile(array_data, q, axis=axis)



### MODULE TO HANDLE A COMPANY PERFORMACE APPRAISAL

class CompanyReturnAppraisal:

    def dupont_analysis(self, float_ni, float_net_revenue, float_avg_ta, float_avg_te, 
                        float_ebt, float_ebit):
        '''
        REFERENCES: https://www.wallstreetprep.com/knowledge/dupont-analysis-template/
        DOCSTRING: DUPONT ANALYSIS IN THREE AND FIVE-STEPS
        INPUTS: NET INCOME, NET REVENUE, AVERAGE TOTAL ASSETS, AVERAGE TOTAL SHAREHOLDER'S EQUITY, 
            EARNS BEFORE TAXES (EBT), EARNS BEFORE EARNS, INTERESTS AND TAXES (EBIT) - IN A 
            5-STEP DUPONT ANALYSIS, THE MULTIPLYING FACTORS ARE:
            1. TAX BRUDEN (NI / EBT) --> THE LOWER RATIO, THE HIGHER THE BURDEN IS
            2. INTEREST BURDEN (EBT / EBIT) --> THE LOWER RATIO, THE HIGHER THE BURDEN IS
            3. OPERATING MARGIN (EBIT / NET REVENUE)
            4. ASSET TURNOVER (NET REVENUE / AVERAGE TOTAL ASSETS)
            5. FINANCIAL LEVERAGE RATIO (AVERAGE TOTAL ASSETS / AVERAGE SHAREHOLDER'S EQUITY)
        OUTPUTS: DICITONARY WITH 3-STEPS AND 5-STEPS DUPONT ROE
        '''
        return {
            '3_steps_dupont_roe': float_ni / float_net_revenue * float_net_revenue / float_avg_ta \
                * float_avg_ta / float_avg_te, 
            '5_steps_dupont_roe': float_ni / float_ebt * float_ebt / float_ebit \
                * float_ebit / float_net_revenue * float_net_revenue / float_avg_ta \
                * float_avg_ta / float_avg_te
        }


from unittest import TestCase
import requests_mock
import urllib.parse

from .fixtures import TOKEN

from typeform import Typeform
from typeform.constants import API_BASE_URL


class ResponsesTestCase(TestCase):
    def setUp(self):
        self.forms = Typeform(TOKEN).forms
        self.responses = Typeform(TOKEN).responses
        form = self.forms.create({
            'title': 'title'
        })
        self.formID = form.get('id')

    def tearDown(self):
        list = self.forms.list()
        forms = list.get('items', [])
        for form in forms:
            self.forms.delete(form.get('id'))

    def test_list_returns_method_and_path(self):
        """
        get all responses has the correct method and path
        """
        with requests_mock.mock() as m:
            url = API_BASE_URL+'/forms/'+self.formID+'/responses'
            m.get(url, json={})
            self.responses.list(self.formID)

            history = m.request_history
            self.assertEqual(history[0].url, url)
            self.assertEqual(history[0].method, 'GET')

    def test_list_correct_params(self):
        """
        paramters are sent correctly
        """
        with requests_mock.mock() as m:
            url = API_BASE_URL+'/forms/'+self.formID+'/responses'
            m.get(url, json={})
            self.responses.list(
                self.formID, pageSize='100', since='2000-01-01T00:00:00Z', completed=True, fields=['1', '2']
            )

            history = m.request_history
            query = history[0].url.split('?')[1]
            params = dict(urllib.parse.parse_qs(query))

            self.assertEqual(params.pop('page_size')[0], '100')
            self.assertEqual(params.pop('since')[0], '2000-01-01T00:00:00Z')
            self.assertEqual(params.pop('completed')[0], 'true')
            self.assertEqual(params.pop('fields')[0], '1,2')

    def test_list_fetches_responses(self):
        """
        get all responses does not throw an error
        """
        result = self.responses.list(self.formID)
        self.assertEqual(result.pop('total_items'), 0)

    def test_list_fetches_responses_with_params(self):
        """
        get all responses does not throw an error with paramters
        """
        result = self.responses.list(
            self.formID, pageSize=100, since='2000-01-01T00:00:00Z', completed=True, fields=['1', '2']
        )
        self.assertEqual(result.pop('total_items'), 0)

    def test_delete_one_token_returns_method_and_path(self):
        """
        delete response has the correct method and path when deleting one token
        """
        with requests_mock.mock() as m:
            url = API_BASE_URL+'/forms/'+self.formID+'/responses?included_tokens=1'
            m.delete(url, json={})
            self.responses.delete(self.formID, includedTokens='1')

            history = m.request_history
            self.assertEqual(history[0].url, url)
            self.assertEqual(history[0].method, 'DELETE')

    def test_delete_multiple_token_returns_method_and_path(self):
        """
        delete response has the correct method and path when deleting multiple tokens
        """
        with requests_mock.mock() as m:
            url = API_BASE_URL+'/forms/'+self.formID+'/responses?included_tokens=1%2C2%2C3'
            m.delete(url, json={})
            self.responses.delete(self.formID, includedTokens=['1', '2', '3'])

            history = m.request_history
            self.assertEqual(history[0].url, url)
            self.assertEqual(history[0].method, 'DELETE')



### LIBRARY TO COLLECT MACROECONOMIC INFORMATIONS FROM BRAZIL ###


import datetime
import backoff
import yfinance as yf
import pandas as pd
from requests import request
from stpstone.settings._global_slots import YAML_BR_MACRO
from stpstone.handling_data.object import HandlingObjects
from stpstone.cals.handling_dates import DatesBR
from stpstone.handling_data.numbers import NumHandler
from stpstone.handling_data.json import JsonFiles
from stpstone.handling_data.html import HtmlHndler
from stpstone.handling_data.str import StrHandler
from stpstone.loggs.db_logs import DBLogs


class Sidra:

    @property
    def rl_dts_indicators(self, str_pool_name=None):
        '''
        REFERENCES: https://www.ibge.gov.br/calendario-indicadores-novoportal.html
        DOCSTRING: RELEASE DATES FROM MACRO SHORT-TERM INDICATORS
        INPUTS: -
        OUTPUTS: DATAFRAME
        '''
        # setting variables
        list_ser = list()
        # request html
        bs_html = HtmlHndler().html_bs_parser(
            YAML_BR_MACRO['sidra']['st_indicators']['url'], 
            bl_verify=YAML_BR_MACRO['sidra']['st_indicators']['bl_verify']
        )
        # looping within short-term indicators raw
        for el in [el.get_text() for el in bs_html.find_all('p')]:
            if StrHandler().match_string_like(el, '*' + YAML_BR_MACRO['sidra'][
                'st_indicators']['pool_names_like'] + '*') == True:
                str_pool_name = StrHandler().get_string_until_substr(
                    el, YAML_BR_MACRO['sidra']['st_indicators']['pool_names_like'])
            elif (str_pool_name != None) and len(el) > 1 and (
                StrHandler().match_string_like(
                    el, '*' + YAML_BR_MACRO['sidra']['st_indicators']['pool_ref_dt'] + '*') == True):
                list_ser.append({
                    YAML_BR_MACRO['sidra']['st_indicators']['col_pool_name']: str_pool_name,
                    YAML_BR_MACRO['sidra']['st_indicators']['col_release_dt']: \
                        DatesBR().str_date_to_datetime(
                            StrHandler().get_string_until_substr(el, 'Refer'), 
                            YAML_BR_MACRO['sidra']['st_indicators']['format_input_dt']
                        ),
                    YAML_BR_MACRO['sidra']['st_indicators']['col_month_calc']: \
                        StrHandler().get_string_after_substr(el, ': ')
                })
        # creating a dataframe from the serialized list
        df_rl_dts_macro = pd.DataFrame(list_ser)
        # id duplicates
        ds_id_dup = df_rl_dts_macro.duplicated(subset=[
            YAML_BR_MACRO['sidra']['st_indicators']['col_pool_name']], keep='last')
        # invert the duplicate strategy and keep all repeated data, but the last
        df_rl_dts_macro = df_rl_dts_macro[ds_id_dup]
        # adding logging
        df_rl_dts_macro = DBLogs().audit_log(
            df_rl_dts_macro, YAML_BR_MACRO['sidra']['st_indicators']['url'], 
            DatesBR().curr_date
        )
        # return dataframe
        return df_rl_dts_macro

    def sidra_variables(self, aggregate, variable='all', location='BR', method='GET',
                        url='https://servicodados.ibge.gov.br/api/v3/agregados/{}/variaveis/{}'):
        '''
        REFERENCES: https://servicodados.ibge.gov.br/api/docs/agregados?versao=3
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # payload
        dict_payload = {
            'localidades': location
        }
        # converting aggregate and variable to string
        aggregate, variable = str(aggregate), str(variable)
        # requesting data
        resp_req = request(method, url.format(aggregate, variable), data=dict_payload)
        # raises exception when not a 2xx response
        resp_req.raise_for_status()
        # retrieve json
        return resp_req.json()

    def sidra_modification_dates(self, aggregate, method='GET',
                                 url='https://servicodados.ibge.gov.br/api/v3/agregados/{}/periodos'):
        '''
        REFERENCES: https://servicodados.ibge.gov.br/api/docs/agregados?versao=3
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # converting aggregate and variable to string
        aggregate = str(aggregate)
        # requesting data
        resp_req = request(method, url.format(aggregate))
        # raises exception when not a 2xx response
        resp_req.raise_for_status()
        # retrieve json
        return resp_req.json()

    @property
    def ipca_modification_dates(self, aggregate='1737'):
        '''
        REFERENCES: https://sidra.ibge.gov.br/tabela/1737
        DOCSTRING: INDICE NUMBER OF IPCA (INFLATION INDEX FOR BRAZILLIAN PURPOSES)
        INPUTS: -
        OUTPUTS: DICTIONARY (CLOSING MONTH AND INDICE NUMBER KEYS)
        '''
        return self.sidra_modification_dates(aggregate)

    @property
    def ipca_15_modification_dates(self, aggregate='3065'):
        '''
        REFERENCES: https://sidra.ibge.gov.br/tabela/3065
        DOCSTRING: INDICE NUMBER OF IPCA 15 (INFLATION INDEX FOR BRAZILLIAN PURPOSES)
        INPUTS: -
        OUTPUTS: DICTIONARY (CLOSING MONTH AND INDICE NUMBER KEYS)
        '''
        return self.sidra_modification_dates(aggregate)

    @property
    def ipca_in_index_number(self, aggregate='1737', variable='2266'):
        '''
        REFERENCES: https://sidra.ibge.gov.br/tabela/1737
        DOCSTRING: INDICE NUMBER OF IPCA (INFLATION INDEX FOR BRAZILLIAN PURPOSES)
        INPUTS: -
        OUTPUTS: DICTIONARY (CLOSING MONTH AND INDICE NUMBER KEYS)
        '''
        return self.sidra_variables(aggregate, variable)[0]['resultados'][0]['series'][0]['serie']

    @property
    def ipca_15_in_index_number(self, aggregate='3065', variable='1117'):
        '''
        REFERENCES: https://sidra.ibge.gov.br/tabela/3065
        DOCSTRING: INDICE NUMBER OF IPCA 15 (INFLATION INDEX FOR BRAZILLIAN PURPOSES)
        INPUTS: -
        OUTPUTS: DICTIONARY (CLOSING MONTH AND INDICE NUMBER KEYS)
        '''
        return self.sidra_variables(aggregate, variable)[0]['resultados'][0]['series'][0]['serie']


class BCB:

    @backoff.on_exception(
        backoff.constant,
        SyntaxError,
        interval=10,
        max_tries=20,
    )
    def sgs_bcb(self, int_sgs_tabel, data_inic, data_fim):
        '''
        REFERENCES: https://dadosabertos.bcb.gov.br/dataset/20542-saldo-da-carteira-de-credito-com-recursos-livres---total/resource/6e2b0c97-afab-4790-b8aa-b9542923cf88
        DOCSTRING: INFOS FROM SGS BCB REST
        INPUTS: INPUTS, INCIAL AND END DATES ('DD/MM/YYYY' OR DATETIME FORMAT)
        OUPUTS: SGS RESPONSE
        '''
        # adjusting variables types
        if isinstance(data_inic, datetime.date) == True:
            data_inic = DatesBR().datetime_to_string(data_inic)
        if isinstance(data_fim, datetime.date) == True:
            data_fim = DatesBR().datetime_to_string(data_fim)
        # url and dict_payload to communicate to rest
        url = 'http://api.bcb.gov.br/dados/serie/bcdata.sgs.{}/dados'.format(int_sgs_tabel)
        query_params = {
            'formato': 'json',
            'dataInicial': data_inic,
            'dataFinal': data_fim
        }
        # retorno do dado de interesse
        message_content = request('GET', url, params=query_params).content
        return HandlingObjects().literal_eval_data(message_content, "b'", "'")

    def igpm(self, data_inic, data_fim, tabela_sgs_bcb='189'):
        '''
        REFERENCES: https://www3.bcb.gov.br/sgspub/consultarvalores/telaCvsSelecionarSeries.paint
        DOCSTRING: IGPM TIMES SERIES
        INPUTS: INICIAL DATE, FINAL DATE, AND SGS BCB TABLE
        OUTPUTS: LIST OF DICTS (DATE, VALUE)
        '''
        return self.sgs_bcb(tabela_sgs_bcb, data_inic, data_fim)

    def selic_daily(self, data_inic, data_fim, tabela_sgs_bcb='11'):
        '''
        REFERENCES: https://www3.bcb.gov.br/sgspub/consultarvalores/telaCvsSelecionarSeries.paint
        DOCSTRING: DAILY SELIC
        INPUTS: INICIAL DATE, FINAL DATE, AND SGS BCB TABLE
        OUTPUTS: LIST OF DICTS (DATE, VALUE)
        '''
        return self.sgs_bcb(tabela_sgs_bcb, data_inic, data_fim)

    def selic_target(self, data_inic, data_fim, tabela_sgs_bcb='432'):
        '''
        REFERENCES: https://www3.bcb.gov.br/sgspub/consultarvalores/telaCvsSelecionarSeries.paint
        DOCSTRING: DAILY SELIC
        INPUTS: INICIAL DATE, FINAL DATE, AND SGS BCB TABLE
        OUTPUTS: LIST OF DICTS (DATE, VALUE)
        '''
        return self.sgs_bcb(tabela_sgs_bcb, data_inic, data_fim)

    def usd_brl(self, data_ref, formato_data_saida='%m/%d/%Y', int_dol_usd_security_id=9800508,
                key_dol_usd_security_id='securityIdentificationCode', key_value='value',
                key_last_update='lastUpdate', int_precision_decimals=4,
                endpoint_taxas_referenciais_b3='https://sistemaswebb3-derivativos.b3.com.br/financialIndicatorsProxy/ReferenceExchangeRate/GetReferenceExchangeRate/eyJsYW5ndWFnZSI6InB0LWJyIn0=',
                method_req='GET', bl_verify=False):
        '''
        REFERENCES: https://olinda.bcb.gov.br/olinda/servico/PTAX/versao/v1/aplicacao#!/recursos/CotacaoDolarDia#eyJmb3JtdWxhcmlvIjp7IiRmb3JtYXQiOiJqc29uIiwiJHRvcCI6MTAwLCJkYXRhQ290YWNhbyI6IjEyLTA5LTIwMjAifSwicHJvcHJpZWRhZGVzIjpbMCwxLDJdfQ==
        DOCSTRING: COMMERCIAL DOL/BRL IN DAILY OBSERVATIONS
        INPUTS: DATE (WHETHER DATETIME OR 'DD/MM/YYYY')
        OUTPUTS: JSON
        '''
        # dateformat adjustment
        if DatesBR().check_date_datetime_format(data_ref) == True:
            data_ref = DatesBR().datetime_to_string(data_ref, formato_data_saida)
        else:
            data_ref = DatesBR().datetime_to_string(DatesBR().str_date_to_datetime(data_ref),
                                                    formato_data_saida)
        # verifying the consistence of data returned from BCB API; in 11/05/2021 there was a
        #   maintenance causing a temporarily unavailability, due to this problem was created a
        #   contingency to return the DOLBRL exchange from B3
        try:
            # request usd/brl
            url = "https://olinda.bcb.gov.br/olinda/servico/PTAX/versao/v1/odata/CotacaoDolarDia(" \
                + "dataCotacao=@dataCotacao)?@dataCotacao={}&$top=100&$format=json&$".format(
                    "'" + data_ref + "'") \
                + "select=cotacaoCompra,cotacaoVenda,dataHoraCotacao"
            dict_payload = {}
            headers = {
                'Cookie': 'JSESSIONID=0000GNzWCNvJ8ltUKMUmrwlH_Y6:1cn7jtfnj; dtCookie=9C0BF4FB'
                + 'DA3E96A68FC297DBEDAE82BF|cHRheHwx; BIGipServer~App~upstream_was_ssl-p='
                + '1020268972.47873.0000'
            }
            request_message = request(
                "GET", url, headers=headers, data=dict_payload).content
            request_message = HandlingObjects().literal_eval_data(request_message, "b'", "'")
            return {
                'cotacao_compra': request_message['value'][0]['cotacaoCompra'],
                'cotacao_venda': request_message['value'][0]['cotacaoVenda'],
                'timestamp_cotacao': request_message['value'][0]['dataHoraCotacao'],
            }
        except:
            # consult the html from b3 reference interest taxes
            resp_req = request(
                method_req, endpoint_taxas_referenciais_b3, verify=bl_verify)
            # raise exception in case return is different from 2xx
            resp_req.raise_for_status()
            # returning content
            json_ref_taxes_b3 = resp_req.content
            # turning type from byte to dictionary
            json_ref_taxes_b3 = HandlingObjects().literal_eval_data(
                json_ref_taxes_b3, "b'", "'")
            # looping through dictionaries and identifying the one that refers to 1-day usd-brl
            dict_usd_brl = [dict_ for dict_ in json_ref_taxes_b3 if int(dict_[
                key_dol_usd_security_id]) == int(int_dol_usd_security_id)][0]
            # return the contingency dolbrl exchange
            return {
                'cotacao_compra': NumHandler().convert_thousands_decimals_separator(
                    dict_usd_brl[key_value], precision_decimals=int_precision_decimals),
                'cotacao_venda': NumHandler().convert_thousands_decimals_separator(
                    dict_usd_brl[key_value], precision_decimals=int_precision_decimals),
                'timestamp_cotacao': dict_usd_brl[key_last_update],
            }

    def foreign_exchange_bcb(self, int_working_days_before_end=1, int_working_days_before_begin=None,
                             url='https://olinda.bcb.gov.br/olinda/servico/PTAX/versao/v1/odata/{}',
                             api_currencies='Moedas',
                             api_currecy_exchange_day='CotacaoMoedaDia(moeda=@moeda,dataCotacao=@dataCotacao)?{}',
                             api_curency_exchange_period='CotacaoMoedaPeriodo(moeda=@moeda,dataInicial=@dataInicial,'
                             + 'dataFinalCotacao=@dataFinalCotacao)?{}', method='GET',
                             key_value='value', key_simbolo='simbolo', str_currency='{}BRL'):
        '''
        REFERENCES: https://olinda.bcb.gov.br/olinda/servico/PTAX/versao/v1/swagger-ui3#/
        DOCSTRING: BRAZILLIAN FOREIGN EXCHANGE FROM ALL CURRENCIES AVAILABE IN OLINDA API FROM
            DOMESTIC CENTRAL BANK
        INPUTS: URL
        OUTPUTS:
        '''
        # setting variables
        dict_export = dict()
        # creating reference datetime, in specific string type, according to working days before
        #   (0 as default)
        datetime_end_ref = DatesBR().sub_working_days(DatesBR().curr_date,
                                                      int_working_days_before_end)
        str_datetime_end_ref = datetime_end_ref.strftime('%m-%d-%Y')
        if int_working_days_before_begin != None:
            datetime_begin_ref = DatesBR().sub_working_days(DatesBR().curr_date,
                                                            int_working_days_before_end)
            str_datetime_begin_ref = datetime_begin_ref.strftime('%m-%d-%Y')
        # header
        dict_header = {
            'accept': 'application/json;odata.metadata=minimal'
        }
        # payload for currencies available consulting
        dict_payload = {
            'format': 'json'
        }
        dict_payload = JsonFiles().dict_to_json(dict_payload)
        # retriving data
        resp_req = request(method, url.format(api_currencies), headers=dict_header,
                           data=dict_payload)
        # raises exception when not a 2xx response
        resp_req.raise_for_status()
        # retrieving json
        json_currencies = resp_req.json()
        # collecting all currencies available (against BRL)
        list_ser_currencies = json_currencies[key_value]
        list_currencies = [str(dict_[key_simbolo])
                           for dict_ in list_ser_currencies]
        # requesting current foreign available for all the currencies disposed before
        for currency in list_currencies:
            #   validating wheter the user aim for a bucket of currencies in a given day or period
            if int_working_days_before_begin == None:
                #   payload for currencies available consulting
                dict_payload = {
                    'format': 'json',
                }
                dict_params = {
                    'moeda': "'{}'".format(str(currency)),
                    'dataCotacao': "'{}'".format(str(str_datetime_end_ref)),
                }
                #   coverting to string payload and params
                str_payload = '%40' + '&%40'.join(
                    '{}={}'.format(k, v) for k, v in 
                    dict_params.items()) + '&%24' + '%24'.join(
                        '{}={}'.format(k, v) for k, v in dict_payload.items())
                #   requesting currency exchange against brl
                resp_req = request(method, url.format(api_currecy_exchange_day.format(
                    str_payload)),
                    headers=dict_header)
            else:
                #   payload for currencies available consulting
                dict_payload = {
                    'format': 'json',
                }
                dict_params = {
                    'moeda': "'{}'".format(str(currency)),
                    'dataInicial': "'{}'".format(str(str_datetime_begin_ref)),
                    'dataFinalCotacao': "'{}'".format(str(str_datetime_end_ref)),
                }
                #   coverting to string payload and params
                str_payload = '%40' + '&%40'.join(
                    '{}={}'.format(k, v) for k, v in
                    dict_params.items()) + '&%24' + '%24'.join(
                        '{}={}'.format(k, v) for k, v in dict_payload.items())
                #   requesting currency exchange against brl
                resp_req = request(method, url.format(api_curency_exchange_period.format(
                    str_payload)), headers=dict_header)
            #   raises exception when not a 2xx response
            resp_req.raise_for_status()
            #   retrieving json
            json_currency_exchange = resp_req.json()
            #   appending to export dictionary
            dict_export[str_currency.format(
                currency)] = json_currency_exchange[key_value]
        # retrieving foreign exchanges
        return dict_export

    @property
    def market_macro_expec(self, url='https://olinda.bcb.gov.br/olinda/servico/Expectativas/versao/v1/odata/ExpectativasMercadoAnuais?$top=100000&$orderby=Data%20desc&$format=json&$select=Indicador,IndicadorDetalhe,Data,DataReferencia,Media,Mediana,Minimo,Maximo,numeroRespondentes', 
        col_indicator='Indicador', col_detailed_indicator='IndicadorDetalhe', col_date='Data', col_ref_year='DataReferencia', 
        col_avg='Media', col_median='Mediana', col_min='Minimo', col_max='Maximo', col_num_answ='numeroRespondentes'):
        '''
        DOCSTRING: ANNUAL MARKET EXPECTATIONS FROM THE BRAZILIAN CENTRAL BANK, INCLUDING INDICATORS LIKE GDP 
            GROWTH AND INFLATION, WITH STATISTICS ON AVERAGE, MEDIAN, MIN, MAX VALUES, AND RESPONDENT COUNT.
        INPUTS: -
        OUTPUTS: DATAFRAME
        '''
        # request olinda bcb
        dict_headers = {
            'accept': 'application/json',
            'client_id': 'E0Bbo4L19nlx',
            'access_token': 's56HuH4yFasr',
            'Cookie': 'BIGipServer~was_p_as3~was_p~pool_was_443_p=4275048876.47873.0000; JSESSIONID=0000X4IrBKiAUyQvbYXXFfX0gne:1dof89mke; TS013694c2=012e4f88b3c6fee6e3a792e5d4f68cb31972d27ba778ec1e05a622b5b87ecf0bda522fe8652f85210b7cbe2b227fe76a647ca3acc6'
        }
        resp_req = requests('GET', url, headers=dict_headers)
        resp_req.raise_for_status()
        json_bcb_expec = resp_req.json()
        # load to pandas dataframe
        df_expec_bcb = pd.DataFrame(json_bcb_expec['value'])
        # changing columns types
        df_expec_bcb = df_expec_bcb.astype({
            col_indicator: str,
            col_detailed_indicator: str,
            col_date: str,
            col_ref_year: int,
            col_avg: float,
            col_median: float,
            col_min: float,
            col_max: float,
            col_num_answ: int
        })
        # return dataframe
        return df_expec_bcb


class YFinanceMacroBR:

    @property
    def ipca_forecast(self, method='GET', url='https://sbcharts.investing.com/events_charts/eu/1165.json',
                      int_convert_miliseconds_seconds=1000):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # requesting data
        req_resp = request(method, url)
        # raises exception when not a 2xx response
        req_resp.raise_for_status()
        # retrieving json
        json_brazillian_cpi = req_resp.json()
        # getting historical data
        json_brazillian_cpi = [{
            'datetime': DatesBR().timestamp_float_to_datetime(
                int(int(dict_['timestamp']) / int_convert_miliseconds_seconds), bl_format=True),
            'actual_state': str(dict_['actual_state']),
            'actual': float(dict_['actual']),
            'forecast': dict_['forecast'],
            'revised': dict_['revised'],
        } for dict_ in json_brazillian_cpi['attr']]
        # retrieving historical data
        return json_brazillian_cpi

    def yforex(self, list_xcg_curr, wd_start_date=2, wd_end_date=0):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # setting variables
        list_ser = list()
        # creating dates of interest according to working days provided
        inf_date = DatesBR().sub_working_days(DatesBR().curr_date, wd_start_date).strftime(
            '%Y-%m-%d')
        sup_date = DatesBR().sub_working_days(DatesBR().curr_date, wd_end_date).strftime(
            '%Y-%m-%d')
        # looping within exchange currencies
        for xcg_curr in list_xcg_curr:
            #   dealing with raw exchange rate currency
            ticker = xcg_curr + "=X"
            #   importing historical data to dataframe
            df_ = yf.download(ticker, start=inf_date, end=sup_date, progress=False)
            #   adding currency column
            df_[YAML_BR_MACRO['exchange_rates_yahoo']['col_currency']] = xcg_curr
            #   adding reference date
            df_[YAML_BR_MACRO['exchange_rates_yahoo']['col_date'].upper()] = \
                df_.index.to_frame(name=YAML_BR_MACRO['exchange_rates_yahoo']['col_date'])
            #   appending to serialized list
            list_ser.extend(df_.to_dict(orient='records'))
        # serialized list to dataframe
        df_xcg = pd.DataFrame(list_ser)
        # renaming columns
        df_xcg = df_xcg.rename(columns={
            x: x.upper() for x in df_xcg.columns
        })
        # sort dataframe
        df_xcg.sort_values(
            [
                YAML_BR_MACRO['exchange_rates_yahoo']['col_currency'], 
                YAML_BR_MACRO['exchange_rates_yahoo']['col_date'].upper()
            ],
            ascending=[True, False],
            inplace=True
        )
        # adding logging to the last dataframe
        df_xcg = DBLogs().audit_log(
            df_xcg, 
            r'https://finance.yahoo.com/quote/',
            DatesBR().utc_from_dt(DatesBR().curr_date)
        )
        # return dataframe
        return df_xcg


### CLASS TO DEAL WITH MICROSOFT CMD TERMINAL ###

class BColors:
    '''
    REFERENCES: https://stackoverflow.com/questions/4406389/if-else-in-a-list-comprehension
    DOCSTRING:
    INPUTS:
    OUTPUTS
    '''
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'



### CHARTS FOR RISK MANAGEMENT ###

import matplotlib.pyplot as plt
from stpstone.finance.b3.infos_fechamento import B3Fcto


class StockMarket:

    def stock_price_evolution(self, list_stocks_tickers, date_inf, date_sup, country, str_x_label,
                              str_y_label):
        '''
        DOCSTRING: CHART OF STOCK PRICE EVOLUTION
        INPUTS: LIST OF STOCKS TICKERS, INFERIOR AND SUPERIOR DATES, COUNTRY 
            (EX.: BRAZIL/ UNITED STATES), STR WITH X LABEL, AND ITS Y LABEL AXIS NAMES 
            FOR THE CHART
        OUTPUTS:
        '''
        # define plt style
        plt.style.use('fivethirtyeight')

        # close price of stocks in portfolio for a given range of dates (n-observations)
        dict_stocks_market_data = dict()
        for k, v in B3Fcto().cotacoes_serie_historica(list_stocks_tickers, date_inf,
                                                      date_sup, country).items():
            list_current_dates = [dict_current_price_data['date'] for dict_current_price_data
                                  in v['historical']]
            list_current_prices = [dict_current_price_data['close'] for dict_current_price_data
                                   in v['historical']]
            dict_stocks_market_data[k] = {
                'dates': list_current_dates,
                'close': list_current_prices
            }

        # plot chart of stock market evoulution through range of time
        plt.figure(figsize=(14, 7))
        for k, v in dict_stocks_market_data.items():
            plt.plot(v['dates'], v['close'], lw=3, alpha=0.8, label=k)
        plt.legend(loc='upper left', fontsize=12)
        plt.xlabel(str_x_label)
        plt.ylabel(str_y_label)
        plt.show()


# StockMarket().stock_price_evolution(['AAPL', 'AMZN', 'GOOGL', 'FB'], '01/01/2016',
#                                     '31/12/2017', 'United States', 'dates', 'price in USD')

# StockMarket().stock_price_evolution(['PETR4', 'PETR3', 'MGLU3', 'VALE3'], '01/01/2018',
#                                     '04/09/2020', 'brazil', 'dates', 'price in BRL')



### ENDPOINT PARA CONEXÃO COM O SITE DADOS DE MERCADO ###

import requests
from pprint import pprint


class DadosDeMercado:
    '''
    REFERENCES: DADOS DE MERCADO ENPOINT:https://www.dadosdemercado.com.br/api/docs/autenticacao
    DOCSTRING: STOCK MARKET DATA FROM B3
    INPUTS: TOKEN
    OUTPUTS: CLASS
    '''

    def __init__(self, token):
        self.token = token

    def fetch_data(self, app, payload=dict(), version_endpoint='1',
                   endpoint='https://api.dadosdemercado.com.br/v{}/{}', method='GET'):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        headers = {
            'Authorization': 'Bearer {}'.format(self.token)
        }
        return requests.request(method, url=endpoint.format(version_endpoint, app),
                                headers=headers, data=payload).json()

    def companies_trading_b3_stock_market(self, app='companies'):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        return self.fetch_data(app)

    def tickers_companies_trading_b3_stock_market(self, app='tickers'):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        return self.fetch_data(app)

    def stock_market_indexes(self, app='indexes'):
        '''
        DOCSTRING: STOCK MARKET INDEXES - NOT PROPERLY WORKING
        INPUTS: APP (INDEXES AS DEFAULT)
        OUTPUTS: JSON
        '''
        return self.fetch_data(app)

    def company_registration_data(self, cvm_code, app='/companies/{}/tickers'):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        if type(cvm_code) != str:
            cvm_code = str(cvm_code)
        return self.fetch_data(app.format(cvm_code))

    def dividends(self, cvm_code, app='/companies/{}/dividends'):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        if type(cvm_code) != str:
            cvm_code = str(cvm_code)
        return self.fetch_data(app.format(cvm_code))

    def splits_inplits(self, cvm_code, app='/companies/{}/splits'):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        if type(cvm_code) != str:
            cvm_code = str(cvm_code)
        return self.fetch_data(app.format(cvm_code))

    def quotes(self, ticker, app='/tickers/{}/quotes'):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        return self.fetch_data(app.format(ticker))

    def yield_curves(self, app='/macro/yield_curves'):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        return self.fetch_data(app)


# # inputs
# token = 'ffc65ca86d9edb68a7ba464b691c1614'
# class_dados_mercado = DadosDeMercado(token)
# cod_cvm = '14311'
# ticker = 'CPLE6'

# # outputs
# json_companies_b3 = class_dados_mercado.companies_trading_b3_stock_market()
# print(json_companies_b3)
# json_tickers_b3 = class_dados_mercado.tickers_companies_trading_b3_stock_market()
# print(json_tickers_b3)
# json_indexes_b3 = class_dados_mercado.stock_market_indexes()
# print(json_indexes_b3)
# json_companies_b3 = class_dados_mercado.company_registration_data(cod_cvm)
# print(json_companies_b3)
# json_dividendos_b3 = class_dados_mercado.dividends(cod_cvm)
# pprint(json_dividendos_b3)
# json_splits_b3 = class_dados_mercado.splits_inplits(cod_cvm)
# print(json_splits_b3)
# json_quotes = class_dados_mercado.quotes(ticker)
# print(json_quotes)
# json_yield_curves = class_dados_mercado.yield_curves()
# print(json_yield_curves)



### EARNINGS MANIPULATION AUDITING

class EarningsManipulation:

    def inputs_beneish_model(self, float_ar_t, float_sales_t, float_ar_tm1, float_sales_tm1, 
                             float_gp_tm1, float_gp_t, float_ppe_t, float_ca_t, float_lti_t, 
                             float_lti_tm1, float_ta_t, float_ppe_tm1, float_ca_tm1, float_ta_tm1, 
                             float_dep_tm1, float_dep_t, float_sga_t, float_sga_tm1, 
                             float_inc_cont_op, float_cfo_t, float_tl_t, float_tl_tm1):
        '''
        REFERENCES: https://corporatefinanceinstitute.com/resources/financial-modeling/beneish-m-score-calculator/
        DOCSTRING: FINANCIAL INPUTS FOR BENEISH M-SCORE MODEL
        INPUTS: ACCOUNTS RECEIVABLES T, SALES T, ACCOUNTS RECEIVABLES T-1, SALES T-1, GROSS PROFIT 
            T-1 (SALES - COGS), GROSS PROFIT T, PPE T, CURRENT ASSETS T, LONG-TERM INVESTMENTS T 
            (SECURITIES), LONG-TERM INVESTMENTS T-1, TOTAL ASSETS T, PPE T-1, CURRENT ASSETS T-1, 
            TOTAL ASSETS T-1, DEPRECIATION T-1, DEPRECIATION T, SG&A T, SG&A T-1, 
            INCOME FROM CONTINUOUS OPERATIONS, CASH FLOW FROM OPERATIONS T
        OUTPUTS: DICTIONARY
        '''
        return {
            'float_dsr': (float_ar_t / float_sales_t) / (float_ar_tm1 / float_sales_tm1), 
            'float_gmi': (float_gp_tm1 / float_sales_tm1) / (float_gp_t / float_sales_t), 
            'float_aqi': (1.0 - (float_ppe_t + float_ca_t + float_lti_t) / float_ta_t) \
                / (1.0 - (float_ppe_tm1 + float_ca_tm1 + float_lti_tm1) / float_ta_tm1), 
            'float_sgi': float_sales_t / float_sales_tm1, 
            'float_depi': (float_dep_tm1 / (float_ppe_tm1 + float_dep_tm1)) / (
                float_dep_t / (float_ppe_t + float_dep_t)
            ), 
            'float_sgai': (float_sga_t / float_sales_t) / (float_sga_tm1 / float_sales_tm1), 
            'float_tata': (float_inc_cont_op - float_cfo_t) / float_ta_t, 
            'float_lvgi': (float_tl_t / float_ta_t )/ (float_tl_tm1 / float_ta_tm1)
        }

    def beneish_model(self, float_dsr, float_gmi, float_aqi, float_sgi, float_depi, float_sgai, 
                      float_tata, float_lvgi):
        '''
        REFERENCES: https://corporatefinanceinstitute.com/resources/financial-modeling/beneish-m-score-calculator/
        DOCSTRING: BENEISH M-SCORE TO CHECK EARNINGS MANIPULATION - Z-SCORE VALUES GREATER THAN 
            -1.78 SHOULD RAISE A WARNING SIGN
        INPUTS: DAYS SALES RECEIVABLES INDEX (DSRI), GROSS MARGING INDEX (GMI), ASSET QUALITY 
            INDEX (AQI), SALES GROWTH INDEX (SGI), DEPRECIATION INDEX (DEPI), SELLING, GENERAL & 
            ADMINISTRATIVE EXPENSES INDEX (SGAI), TOTAL ACCRUALS TO TOTAL ASSETS (TATA), 
            LEVERAGE INDEX (LVGI)
        OUTPUTS: FLOAT
        '''
        return -4.84 + 0.920 * float_dsr + 0.528 * float_gmi + 0.404 * float_aqi \
            + 0.892 * float_sgi + 0.115 * float_depi - 0.172 * float_sgai + 4.679 * float_tata \
            - 0.327 * float_lvgi


from unittest import TestCase
import requests_mock

from .fixtures import TOKEN

from typeform.client import Client
from typeform.constants import API_BASE_URL


class ClientTestCase(TestCase):
    def test_client_pass_headers(self):
        """
        request pass correct headers
        """
        client = Client(TOKEN)

        with requests_mock.mock() as m:
            m.get(API_BASE_URL+'/forms')
            client.request('GET', '/forms', data={}, params={}, headers={
                'Accepts': 'application/json'
            })

            history = m.request_history
            self.assertEqual(history[0].url, API_BASE_URL+'/forms')
            self.assertEqual(history[0].headers.pop('Content-Type'), 'application/json')
            self.assertEqual(history[0].headers.pop('Accepts'), 'application/json')
            self.assertEqual(history[0].headers.pop('Authorization'), 'bearer %s' % TOKEN)



import json
import requests
import typing

from .constants import API_BASE_URL
from .utils import buildUrlWithParams, mergeDict


class Client(object):
    """TypeForm API HTTP client"""

    def __init__(self, token: str, headers: dict = {}):
        """Constructor for TypeForm API client"""
        self.__headers = mergeDict({
            'Content-Type': 'application/json',
            'Accept': 'application/json',
            'Authorization': 'bearer %s' % token
        }, headers)

    def request(self, method: str, url: str, data: any = {}, params: dict = {},
                headers={}) -> typing.Union[str, dict]:
        requestUrl = buildUrlWithParams((API_BASE_URL + url), params)
        requestHeaders = mergeDict(self.__headers, headers)
        requestData = ''
        if type(data) is dict:
            requestData = json.dumps(data) if len(data.keys()) > 0 else ''

        if type(data) is list:
            requestData = json.dumps(data) if len(data) > 0 else ''

        result = requests.request(method, requestUrl, data=requestData, headers=requestHeaders,
                                  verify=False)
        return self.__validator(result)

    def __validator(self, result: requests.Response) -> typing.Union[str, dict]:
        try:
            body = json.loads(result.text)
        except Exception:
            body = {}

        if type(body) is dict and body.get('code', None) is not None:
            raise Exception(body.get('description'))
        elif result.status_code >= 400:
            raise Exception(result.reason)
        elif len(result.text) == 0:
            return 'OK'
        else:
            return body



# PERFORMANCE APPRAISAL FOR FIXED INCOME BILLS, NOTES, BONDS, AND OTHER ASSETS

import numpy as np
from stpstone.finance.derivatives.futures import FuturesPricingB3


class FixedIncomeAppraisal:

    def __init__(self, list_cfs=None, list_rates=None, present_value=None, ytm=None,
                 no_coupons=None, payments_per_year=2, pv_minus=None, pv_plus=None,
                 delta_yield=None, delta_pv=None,
                 delta_curve=None, annual_modified_duration=None, future_value=None,
                 shift=0.0001, bl_exact_form_macaulay_dur=True, bl_exact_form_money_dur=True,
                 bl_exact_form_convexity=True, bl_exact_form_pv_01=True,
                 bl_consider_convexity_delta_pv=True):
        self.list_cfs = list_cfs
        self.list_rates = list_rates
        self.present_value = present_value
        self.ytm = ytm
        self.no_coupons = no_coupons
        self.payments_per_year = payments_per_year
        self.bl_exact_form_macaulay_dur = bl_exact_form_macaulay_dur
        self.pv_minus = pv_minus
        self.pv_plus = pv_plus
        self.delta_yield = delta_yield
        self.delta_pv = delta_pv
        self.delta_curve = delta_curve
        self.annual_modified_duration = annual_modified_duration
        self.future_value = future_value
        self.shift = shift
        self.bl_exact_form_money_dur = bl_exact_form_money_dur
        self.bl_exact_form_convexity = bl_exact_form_convexity
        self.bl_exact_form_pv_01 = bl_exact_form_pv_01
        self.bl_consider_convexity_delta_pv = bl_consider_convexity_delta_pv

    @property
    def yield_durations(self):
        '''
        REFERENCE: https://quant.stackexchange.com/questions/15549/modified-or-macauley-duration-in-python
        DOCSTRING: YIELD DURATIONS - MACAULAY AND MODIFIED
        INPUTS: CASH FLOWS, list_rates (FLOAT OR LIST OF FLOATS), PRESENT VALUE OF BOND, YTM, NUMBER
            OF CUPONS AND PAYMENTS PER YEAR
        OUTPUTS: DICTIONARY WITH MACAULAY_DURATION AND MODIFIED_DURATION
        '''
        # calculating wheter through exact or approximated form
        if self.bl_exact_form_macaulay_dur == False:
            mac_dur = np.sum([self.list_cfs[i] * ((i + 1) / self.payments_per_year) /
                              np.power(1 + self.list_rates[i], i + 1) for i in range(
                                  len(self.list_cfs))]) / \
                self.present_value
        else:
            mac_dur = (self.pv_minus - self.pv_plus) / \
                (2 * self.delta_yield * self.present_value)
        # modified duration
        mod_dur = mac_dur / (1 + self.ytm / self.no_coupons)
        # return durations
        return {
            'macaulay_duration': mac_dur,
            'modified_duration': mod_dur
        }

    @property
    def curve_durations(self):
        '''
        DOCSTRING: CURVE DURATIONS
        INPUTS: PV MINUS, PV PLUS, PV ZERO AND DELTA CURVE
        OUTPUTS:
        '''
        # effective duration
        eff_dur = (self.pv_minus - self.pv_plus) / \
            (2 * self.delta_curve * self.present_value)
        # returning durations
        return {
            'effective_duration': eff_dur
        }

    @property
    def key_rate_duration(self):
        '''
        DOCSTRING: KEY RATE DURATION IS THE BOND SENSITIVITY TO THE BENCHMARK YIELD CURVE IN A 
            SPECIFIC MATURITY, PROVIDING FURTHER SENSITIVITY TO NON-PARALLEL BENCHMARK YIELD CURVE 
            CHANGES
        INPUTS: PV, DELTA PV, DELTA RATE K
        OUTPUTS: FLOAT
        '''
        return - 1.0 / self.present_value * (self.delta_pv / self.delta_yield)

    @property
    def money_duration(self):
        '''
        DOCSTRING: THE MONEY DURATION OF A BOND IS A MEASURE OF THE PRICE CHANGE IN UNITS OF THE 
            CURRENCY IN WHICH THE BOND IS DENOMINATED
        INPUTS: ANNUAL MODIFIED DURATION (FLOAT), AND PV (FLOAT)
        OUTPUTS: FLOAT
        '''
        # calculating wheter through exact or approximated form
        if self.bl_exact_form_money_dur == True:
            return self.yield_durations['macaulay_duration'] * self.present_value
        else:
            return self.annual_modified_duration * self.present_value

    @property
    def convexity(self):
        '''
        REFERENCE: https://quant.stackexchange.com/questions/15549/modified-or-macauley-duration-in-python
        DOCSTRING: BOND CONVEXITY ADJUSTMENT - THE SECOND ORDER EFFECT
        INPUTS: CASH FLOWS, list_rates (FLOAT OR LIST OF FLOATS), PRESENT VALUE OF BOND, YTM, NUMBER
            OF CUPONS AND PAYMENTS PER YEAR
        OUTPUTS: TUPLE OF FLOATS
        '''
        # calculating wheter through exact or approximated form
        if self.bl_exact_form_convexity == True:
            return np.sum([self.list_cfs[i] * ((i + 1) * (i + 2) / self.payments_per_year) /
                           np.power(1 + self.list_rates[i], i + 3) for i in range(
                               len(self.list_cfs))]) / self.future_value
        else:
            return ((self.pv_minus + self.pv_plus) - (2 * self.present_value)) / (
                self.delta_yield ** 2 / self.present_value)

    def dv01(self, ytm, nper, side='C', contract='DI1'):
        '''
        DOCSTRING: DV01 TO MEASURE A DI1 PNL APPLIED A STRESS OF 1 BPS (0.1%) IN YTM
        INPUTS: YTM, NPER, SIDE
        OUPUTS: FLOAT
        '''
        if contract == 'DI1':
            if side == 'C':
                reverse_side = 'V'
            elif side == 'V':
                reverse_side = 'C'
            else:
                raise Exception(
                    'Poorly defined side, ought be "C" or "V", please revisit the parameter.')
            return FuturesPricingB3().pu_di1(ytm, nper, c_v_taxa=side) \
                + FuturesPricingB3().pu_di1(ytm + self.shift, nper, c_v_taxa=reverse_side)
        else:
            raise Exception('Contract poorly definied')

    @property
    def pv01(self):
        '''
        REFERENCES: https://www.theice.com/publicdocs/futures/Price_Sensitivity.pdf
        DOCSTRING: PV01, OR BASIS POINT VALUE, IS THE PRICE SENSITIVITY TO AN YIELD PERCENTUAL 
            CHANGE
        INPUTS: CASH FLOWS, list_rates (FLOAT OR LIST OF FLOATS), PRESENT VALUE OF BOND, YTM, NUMBER
            OF CUPONS AND PAYMENTS PER YEAR (2 AS DEFAULT) AND SHIFT (0.0001 AS DEFAULT), BL 
            DIRECT FORM, PV MINUS (NONE DEFAULT) AND PV PLUS (NONE DEFAULT)
        OUPUTS: FLOAT
        '''
        # calculating wheter through exact or approximated form
        if self.bl_exact_form_pv_01 == True:
            return self.yield_durations(
                self.list_cfs, self.list_rates, self.present_value, self.ytm, self.no_coupons,
                self.payments_per_year)['modified_duration'] * self.present_value * self.shift
        else:
            return (self.pv_minus - self.pv_plus) / 2.0

    @property
    def approximate_delta_bond_price(self):
        '''
        DOCSTRING: ESTIMAGTE PERCENTAGE PRICE CHANGE IN CURRENCY UNITS
        INPUTS: ANNUAL MODIFIED DURATION, PV, DELTA YIELD
        OUTPUTS: FLOAT
        '''
        # calculating with or without second order bond pricing - convexity
        if self.bl_consider_convexity_delta_pv == False:
            return -self.money_duration(self.annual_modified_duration, self.present_value) \
                * self.delta_yield
        else:
            return (-self.yield_durations(
                self.list_cfs, self.list_rates, self.present_value, self.ytm, self.no_coupons,
                self.payments_per_year)['modified_duration'] * self.delta_yield) \
                + (0.5 * self.convexity(self.list_cfs, self.list_rates, self.future_value,
                                        self.payments_per_year) * self.delta_yield ** 2)



from .forms import Forms
from .responses import Responses
from .client import Client

__all__ = ['Typeform']


class Typeform:
    """Typeform API client"""

    def __init__(self, token: str, headers: dict = {}):
        """Constructor for Typeform API client"""
        client = Client(token, headers=headers)
        self.__forms = Forms(client)
        self.__responses = Responses(client)

    @property
    def forms(self) -> Forms:
        return self.__forms

    @property
    def responses(self) -> Responses:
        return self.__responses
